import sys
from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QApplication, QLabel, QTableView, QMessageBox
from PyQt5.QtGui import QIcon, QStandardItemModel, QStandardItem, QColor
import hashlib
import binascii
from service.service import get_server_params, AppConfig
import server_db
from server_db import ServerDB
from reg_new_user import RegNewUser

sys.path.append('../')


class MainApp(QMainWindow):
    """
    Класс описывает основное окно программы GUI-версии серверной части
    """
    def __init__(self, parent=None):
        super().__init__()
        self.initUI()

    def initUI(self):
        self.setFixedSize(600, 400)
        self.setWindowTitle('Server conf for chat')

        self.data_model = None

        # Здесь одна метка и одна таблица для всех случаев. Буду переименовывать по ситуации
        self.label = QLabel('', self)
        self.label.setFixedSize(240, 20)
        self.label.move(10, 50)

        self.table = QTableView(self)
        self.table.move(10, 80)
        self.table.setFixedSize(580, 310)

        exitAction = QAction(QIcon(''), 'Завершить работу', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.triggered.connect(qApp.quit)

        # В интерфейсе можно сделать отдельную кнопку refresh для обновления списка записей.
        # Здесь повторное нажатие на кнопку на панели инструментов вызовет обновление списка
        self.btnAllClients = QAction(QIcon(''), 'All users', self)
        self.btnAllClients.triggered.connect(self.show_all_clients)

        self.btnStat = QAction(QIcon(''), 'Sessions', self)
        self.btnStat.triggered.connect(self.show_sessions)

        self.btnMessages = QAction(QIcon(''), 'Messages', self)
        self.btnMessages.triggered.connect(self.show_messages)

        self.btnServerConf = QAction(QIcon(''), 'Server config', self)
        self.btnServerConf.triggered.connect(self.show_serverconf)

        self.btnRegNewUser = QAction(QIcon(''), 'Reg new user', self)
        self.btnRegNewUser.triggered.connect(self.reg_new_user)

        self.toolbar = self.addToolBar('Toolbar')
        self.toolbar.setFixedHeight(40)
        self.toolbar.addAction(exitAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.btnAllClients)
        self.toolbar.addAction(self.btnStat)
        self.toolbar.addAction(self.btnMessages)
        self.toolbar.addAction(self.btnRegNewUser)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.btnServerConf)

        self.show()

    def show_all_clients(self):
        """
        Метод формирует и отображает содержимое таблицы Users
        :return:
        """
        self.label.setText('All users:')
        self.data_model = QStandardItemModel(self)
        self.data_model.setHorizontalHeaderLabels(['id', 'user name', 'user password'])

        for item in server_db.session.query(server_db.Users).all():
            table_item = []
            table_item.append(QStandardItem(str(item.id)))
            table_item.append(QStandardItem(item.name))
            table_item.append(QStandardItem(str(item.passwd)))
            # Вот тут бред, setEditable устанавливать на каждую ячейку. Почему нет setEditable для всего data_model?
            # Или на худой конец - для отдельной строки или колонки?
            for item in table_item:
                item.setEditable(False)
            self.data_model.appendRow(table_item)

        self.table.setModel(self.data_model)
        self.table.resizeColumnsToContents()

    def show_sessions(self):
        """
        Метод формирует и отображает содержимое таблицы Sessions
        :return:
        """
        self.label.setText('Sessions')

        self.data_model = QStandardItemModel(self)
        self.data_model.setHorizontalHeaderLabels(['id', 'user id', 'start session', 'ip address', 'port', 'end session'])

        for item in server_db.session.query(server_db.Sessions).all():
            table_item = []
            table_item.append(QStandardItem(str(item.id)))
            # Сюда нужно подтянуть user_name
            table_item.append(QStandardItem(str(item.user_id)))
            table_item.append(QStandardItem(str(item.start_session)))
            table_item.append(QStandardItem(item.ip_address))
            table_item.append(QStandardItem(str(item.port)))
            table_item.append(QStandardItem(str(item.end_session)))
            for item in table_item:
                item.setEditable(False)
            self.data_model.appendRow(table_item)

        self.table.setModel(self.data_model)
        self.table.resizeColumnsToContents()

    def show_messages(self):
        """
        Метод формирует и отображает содержимое таблицы Messages
        :return:
        """
        self.label.setText('Messages:')

        self.data_model = QStandardItemModel(self)
        self.data_model.setHorizontalHeaderLabels(['id', 'date time', 'from user id', 'to user id', 'message'])

        for item in server_db.session.query(server_db.Messages).all():
            table_item = []
            table_item.append(QStandardItem(str(item.id)))
            # Сюда нужно подтянуть user_name
            table_item.append(QStandardItem(str(item.mdatetime)))
            table_item.append(QStandardItem(str(item.from_user_id)))
            table_item.append(QStandardItem(str(item.to_user_id)))
            table_item.append(QStandardItem(item.message))
            for item in table_item:
                item.setEditable(False)
            self.data_model.appendRow(table_item)

        self.table.setModel(self.data_model)
        self.table.resizeColumnsToContents()

    def reg_new_user(self):
        """
        Метод вызывает диалоговое окно для регистрации новго пользователя на сервере
        :return:
        """
        global select_dialog
        create_user_dialog = RegNewUser()
        create_user_dialog.btn_ok.clicked.connect(lambda: self.create_new_user(create_user_dialog))
        create_user_dialog.show()

    def create_new_user(self, dialog):
        """
        Метод проверяет корректность данных, введенных в окне регистрации нового пользователя,
        хэширует пароль и сохраняет данные в таблицу Users
        :param dialog: Экземпляр диалогового окна
        :return:
        """
        # Берем из диалога имя пользователя, пароль, подтверждение пароля
        username = dialog.username_edit.text()
        userpasswd = dialog.userpasswd_edit.text()
        passwdconfirm = dialog.passwdconfirm_edit.text()

        if username == '' or userpasswd == '' or passwdconfirm == '':
            QMessageBox.information(self, 'Ошибка!', 'Ошибка при вводе данных! Одно или более полей не заполнены!',
                                    QMessageBox.Ok,
                                    QMessageBox.Ok)
        elif userpasswd != passwdconfirm:
            QMessageBox.information(self, 'Ошибка!', 'Ошибка при вводе данных! Пароль и подтверждение не совпадают!',
                                    QMessageBox.Ok,
                                    QMessageBox.Ok)
            # Вызываем метод, который отправит сообщение на сервер - проверить существование такого пользователя
        elif server_db.check_username(username):
            QMessageBox.information(self, 'Ошибка!', 'Пользователь с указанным именем на сервере уже существует!',
                                    QMessageBox.Ok,
                                    QMessageBox.Ok)
        else:
            # Генерируем хэш пароля, в качестве соли будем использовать логин в
            # нижнем регистре.
            userpasswd = userpasswd.encode('utf-8')
            salt = username.lower().encode('utf-8')
            passwd_hash = hashlib.pbkdf2_hmac('sha512', userpasswd, salt, 100000)
            server_db.reg_new_user(username, binascii.hexlify(passwd_hash))
            QMessageBox.information(self, 'Информация!', 'Пользователь успешно зарегистрирован!',
                                    QMessageBox.Ok,
                                    QMessageBox.Ok)

        # Закрываем диалоговое окно
        dialog.close()

    def show_serverconf(self):
        """
        Метод описывает содержимое окна для ввода параметров приложения
        :return:
        """
        # Функция обработчик открытия окна выбора папки
        # def open_file_dialog():
        #     global dialog
        #     dialog = QFileDialog(self)
        #     path = dialog.getExistingDirectory() # get any file???
        #     path = path.replace('/', '\\')
        #     self.db_path.insert(path)
        # return path

        self.label.setText('Server config:')

        config = AppConfig()
        client_host = config.client_host
        client_port = config.client_port
        client_username = config.client_username
        server_hosts = config.server_hosts
        server_port = config.server_port
        db_filename = config.db_filename

        self.data_model = QStandardItemModel(self)
        self.data_model.setHorizontalHeaderLabels(['param name', 'param value'])

        table_item = []
        cell = QStandardItem('[DATABASE]')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem('')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('db_filename')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem(db_filename)
        cell.setEditable(True)
        table_item.append(cell)
        self.data_model.appendRow(table_item)
        # И вот тут бы отловить click(), чтобы вызвать диалог выбора файла...

        table_item = []
        cell = QStandardItem('[SERVER]')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem('')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('hosts')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem(server_hosts)
        cell.setEditable(True)
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('port')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem(server_port)
        cell.setEditable(True)
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('[CLIENT]')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem('')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('host')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem(client_host)
        cell.setEditable(True)
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('port')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem(client_port)
        cell.setEditable(True)
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        table_item = []
        cell = QStandardItem('username')
        cell.setEditable(False)
        cell.setBackground(QColor('lightGray'))
        table_item.append(cell)
        cell = QStandardItem(client_username)
        cell.setEditable(True)
        table_item.append(cell)
        self.data_model.appendRow(table_item)

        self.table.setModel(self.data_model)
        self.table.setColumnWidth(1, 400)


if __name__ == '__main__':
    hosts, port, db_filename = get_server_params(sys.argv, 'server')
    server_db = ServerDB(db_filename)

    app = QApplication(sys.argv)
    main_app = MainApp()
    sys.exit(app.exec_())



from sqlalchemy import create_engine
from sqlalchemy import Table, Column, Integer, DateTime, String, MetaData, ForeignKey
from datetime import datetime
from sqlalchemy.orm import mapper, sessionmaker
import time, datetime


class ServerDB:
    """
    Класс описывает серверную БД приложения
    Users - таблица со список всех зарегистрированных на сервере клиентов
    Sessions - таблица содержит перечень всех сессий клиентов
    Messages - таблица содержит всю историю сообщений
    """
    class Users:
        def __init__(self, id, name, passwd):
            self.id = id
            self.name = name
            self.passwd = passwd

        def __repr__(self):
            return f'User: {self.name}'

    class Sessions:
        def __init__(self, id, user_id, start_session, ip_address, port, end_session=None):
            self.id = id
            self.user_id = user_id
            self.start_session = start_session
            self.ip_address = ip_address
            self.port = port
            self.end_session = end_session

        def __repr__(self):
            return f'User_id: {self.user_id}, login_datetime: {self.start_session}'

    # Таблица с сообщениями
    class Messages:
        def __init__(self, id, mdatetime, from_user_id, to_user_id, message):
            self.id = id
            self.mdatetime = mdatetime
            self.from_user_id = from_user_id
            self.to_user_id = to_user_id
            self.message = message
            # Тут можно подойти творчески, добавить признаки доставки сообщения, прочтения сообщения,
            # отложенного сообщения, доставки сообщения в строго определенное время, ...

    def __init__(self, db_filename):
        self.sql_engine = None
        try:
            self.sql_engine = create_engine(db_filename, pool_recycle=3600)
        except Exception as e:
            print(e)

        self.metadata = MetaData()

        users = Table('users', self.metadata,
                      Column('id', Integer, primary_key=True),
                      Column('name', String, nullable=False, unique=True),
                      Column('passwd', String, nullable=False)
                      )

        sessions = Table('sessions', self.metadata,
                         Column('id', Integer, primary_key=True),
                         Column('user_id', ForeignKey('users.id')),
                         Column('start_session', DateTime, default=datetime.datetime.now()),
                         Column('ip_address', String),
                         Column('port', Integer),
                         Column('end_session', DateTime)
                         )

        messages = Table('messages', self.metadata,
                         Column('id', Integer, primary_key=True),
                         Column('mdatetime', DateTime, default=datetime.datetime.now()),
                         Column('from_user_id', ForeignKey('users.id')),
                         Column('to_user_id', ForeignKey('users.id')),
                         Column('message', String)
                         )

        self.metadata.create_all(self.sql_engine)

        mapper(self.Users, users)
        mapper(self.Sessions, sessions)
        mapper(self.Messages, messages)

        Session = sessionmaker(bind=self.sql_engine)
        self.session = Session()

    def append_user(self, name_user, passwd):
        """
        Метод регистрирует в таблице Users нового пользователя
        UPD: устаревший метод(!), начиная с Lesson06 не используется (???)
        :param name_user: Имя пользователя
        :param passwd: Схэшированный пароль пользователя
        :return: id пользователя
        """
        # Сначала проверяем, есть ли пользователь в базе?
        id_user = 0
        row_user = self.session.query(self.Users).filter_by(name=name_user).first()
        if row_user is None:
            # Определяем следующий свободный id. Допускаю, что есть более лаконичный вариант, но пока так
            row_user = self.session.query(self.Users.id).order_by(self.Users.id.desc()).first()
            if row_user is None:
                id_user = 1
            else:
                id_user = row_user[0] + 1

            row_user = self.Users(id_user, name_user, passwd)
            self.session.add(row_user)
            self.session.commit()
        else:
            id_user = row_user.id

        return id_user

    def append_session(self, id_user, ip_address, port):
        """
        Метод добавляет сессию пользователя в таблицу Sessions
        :param id_user: Имя пользователя
        :param ip_address: ip-адрес клиента
        :param port: Порт клиента
        :return: id сессии
        """
        # Добавляем сессию
        id_session = self.session.query(self.Sessions.id).order_by(self.Sessions.id.desc()).first()
        if id_session is None:
            id_session = 1
        else:
            id_session = id_session[0] + 1

        row = self.Sessions(id_session, id_user, datetime.datetime.now(), ip_address, port)
        self.session.add(row)
        self.session.commit()

        return id_session # здесь нужно подумать, что возвращать

    # Регистрация нового пользователя из графического интерфейса сервера, без добавления сессии
    def reg_new_user(self, name_user, passwd):
        """
        Метод, добавляющий пользователя в таблицу Users
        :param name_user: Имя пользователя
        :param passwd: Схэшированный пароль
        :return:
        """
        # Сначала проверяем, есть ли пользователь в базе?
        id_user = 0
        row_user = self.session.query(self.Users).filter_by(name=name_user).first()
        if row_user is None:
            # Определяем следующий свободный id. Допускаю, что есть более лаконичный вариант, но пока так
            row_user = self.session.query(self.Users.id).order_by(self.Users.id.desc()).first()
            if row_user is None:
                id_user = 1
            else:
                id_user = row_user[0] + 1

            row_user = self.Users(id_user, name_user, passwd)
            self.session.add(row_user)
            self.session.commit()

    def close_session(self, id_session):
        """
        Метод записывает дату/время окончания сессии
        :param id_session: id сессии
        :return:
        """
        # Сначала проверяем, есть ли пользователь в базе?
        row_session = self.session.query(self.Sessions).filter_by(id=id_session).first()
        if row_session is None:
            # Сессия не найдена. Можно в лог отправить сообщение об ошибке
            pass
        else:
            # Сессия в БД есть
            # Обновляем дату/время завершения сессии
            row_session = self.session.query(self.Sessions).filter_by(id=id_session).update({'end_session': datetime.datetime.now()})
            self.session.commit()

    def append_message(self, from_user_id, to_user_id, message):
        """
        Метод сохраняет в таблицу Messages новое сообщение
        :param from_user_id: id отправителя сообщения
        :param to_user_id: id получателя сообщения
        :param message: Текст сообщения
        :return:
        """
        # Здесь исправить!
        last_id = self.session.query(self.Messages.id).order_by(self.Messages.id.desc()).first()
        if last_id is None:
            last_id = 1
        else:
            last_id = last_id[0] + 1

        row = self.Messages(last_id, datetime.datetime.now(), from_user_id, to_user_id, message)
        self.session.add(row)
        self.session.commit()

    # Возвращаем в виде списка перечень всех пользователей. Можно добавить условие - только активных пользователей
    # def get_users_list(self, is_active=1):
    def get_users_list(self):
        """
        Метод возвращает в виде списка список всех пользователей из таблицы Users
        Метод использовался в CLI-версии программы
        В GUI-версии не используется
        :return: Список пользователей
        """
        users_list = []
        for item in self.session.query(self.Users).all():
            # if is_active:
            #     if item.end_session is None:
            #         users_list.append(item.name)
            # else:
            users_list.append(item.name)

        return users_list

    # Запрашиваем пароль пользователя из БД
    def get_user_passwd(self, username):
        """
        Функция по имени пользователя возвращает схэшированный пароль
        :param username: Имя пользователя
        :return: Пароль
        """
        row_users = self.session.query(self.Users).filter(self.Users.name==username).first()
        if row_users is None:
            return False
        else:
            return row_users.passwd

    def check_username(self, username):
        """
        Метод проверяет существование пользователя на сервере, в случае успеха возвращает его id
        :param username: Имя пользователя
        :return: id пользователя
        """
        row_users = self.session.query(self.Users).filter(self.Users.name==username).first()
        if row_users is None:
            return False
        else:
            return row_users.id

    def get_user_id(self, username):
        """
        Метод по имени пользователя возвращает его id
        :param username: Имя пользователя
        :return: id пользователя
        """
        row_users = self.session.query(self.Users).filter(self.Users.name==username).first()
        if row_users is None:
            return False
        else:
            return row_users.id

    def get_user_session(self, id_user):
        """
        Метод возвращает id активной сессии пользователя
        :param id_user: id пользователя
        :return: id сессии
        """
        # Сначала проверяем, есть ли пользователь в базе?
        row_session = self.session.query(self.Sessions).filter(self.Sessions.user_id==id_user, self.Sessions.end_session == None).first()
        if row_session is None:
            # Сессия не найдена. Можно в лог отправить сообщение об ошибке
            return None
        else:
            # Сессия в БД есть
            return row_session.id

    def save_user_end_session(self, id_user):
        """
        Странный метод
        :param id_user:
        :return:
        """
        pass



from PyQt5.QtWidgets import QDialog, QLabel, QPushButton, QLineEdit
from PyQt5.QtCore import Qt


class RegNewUser(QDialog):
    """
    Класс, описывающий окно для регистрации нового пользователя на сервере
    """
    def __init__(self):
        super().__init__()

        self.setFixedSize(350, 240)
        self.setWindowTitle('Регистрация пользователя')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.username_label = QLabel('Имя пользователя', self)
        self.username_label.setFixedSize(320, 20)
        self.username_label.move(10, 10)

        self.username_edit = QLineEdit(self)
        self.username_edit.setFixedSize(320, 20)
        self.username_edit.move(10, 40)

        self.userpasswd_label = QLabel('Пароль', self)
        self.userpasswd_label.setFixedSize(320, 20)
        self.userpasswd_label.move(10, 70)

        self.userpasswd_edit = QLineEdit(self)
        self.userpasswd_edit.setEchoMode(QLineEdit.Password)
        self.userpasswd_edit.setFixedSize(320, 20)
        self.userpasswd_edit.move(10, 100)

        self.passwdconfirm_label = QLabel('Подтверждение', self)
        self.passwdconfirm_label.setFixedSize(320, 20)
        self.passwdconfirm_label.move(10, 130)

        self.passwdconfirm_edit = QLineEdit(self)
        self.passwdconfirm_edit.setEchoMode(QLineEdit.Password)
        self.passwdconfirm_edit.setFixedSize(320, 20)
        self.passwdconfirm_edit.move(10, 160)

        self.btn_ok = QPushButton('OK', self)
        self.btn_ok.setFixedSize(100, 40)
        self.btn_ok.move(130, 190)

        self.btn_cancel = QPushButton('Отмена', self)
        self.btn_cancel.setFixedSize(100, 40)
        self.btn_cancel.move(240, 190)
        self.btn_cancel.clicked.connect(self.close)



from setuptools import setup, find_packages

setup(name="rtrv_my_chat_server",
      version="1.0.1",
      description="my_chat_server",
      author="RTretyakov",
      author_email="rtretyakov@outlook.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex'],
      scripts=['server/server_run']
      )



import select
import os
import sys
import hmac
import logging
import log.client_log_config
import log.server_log_config
from service.service import decode_msg, get_server_params, encode_msg, create_socket
from decos import log
from serv_descr import ServerDescr
from server_db import ServerDB

SECRET_KEY = b'abrakadabra'


class ServerVerifier(type):
    pass


class Server(metaclass=ServerVerifier):
    """
    Класс описывает серверную сторону приложения
    """
    hosts = ServerDescr('hosts', str, '')
    port = ServerDescr('port', int, 7777)

    def __init__(self):
        self.hosts = hosts
        self.port = port
        self.server_socket = None
        self.sockets = []   # Список сокетов - подключенных клиентов. Сюда добавляются элементы
        # при установлении нового соединения
        self.remote_socket = None
        self.address = None
        self.responses = []
        self.read_sockets = []
        self.write_sockets = []
        self.usersdict = {}  # Словарь вида {username1: socket1, username2: socket2, ...}
        self.user_session = {}
        self.create_socket()

    def create_socket(self):
        """
        Метод создает серверный сокет
        :return:
        """
        self.server_socket = create_socket('server')
        self.server_socket.bind((self.hosts, self.port))
        self.server_socket.listen(5)
        self.server_socket.settimeout(0.1)

    def get_client_socket(self, username):
        """
        Функция по имени пользователя возвращает сокет получателя сообщения
        :param username:  - имя пользователя
        :return:  - сокет пользователя
        """
        if username in self.usersdict.keys():
            return self.usersdict[username]

        log.info(f'В активных сокетах нет клиента с именем {username}')
        return None

    def accept_connection(self):
        """
        Метод определяет попытку подключения к серверу
        Проводит авторизацию подключенного клиентв
        Проводит аутентификацию клиента по имени пользователя и паролю
        :return:
        """
        try:
            self.remote_socket, self.address = self.server_socket.accept()
        except OSError as e:
            pass
        else:
            # Получили азпрос на установление соединения. Проводим авторизацию нового пользователя
            # Генерим тестовое сообщение
            message = os.urandom(32)
            # Отправлям только что подключенному клиенту
            self.remote_socket.send(message)
            # На основе тестового сообщения и секретного ключа генерим хэш
            hash = hmac.new(SECRET_KEY, message, 'MD5')
            digest = hash.digest()
            # Получаем ответный хэш от клиента
            response = self.remote_socket.recv(len(digest))
            # Сравниваем наш хэш с клиентским
            if hmac.compare_digest(digest, response):
                log.info(f'Установлено соединение с {self.remote_socket} {self.address}')
                # Добавялем сокет в список прослушиваемых для дальнейшей работы
                self.sockets.append(self.remote_socket)
                # Оповещаем клиента об успешной авторизации
                self.remote_socket.send('Auth ok!'.encode('utf-8'))

                # Теперь аутентификация
                response = decode_msg(self.remote_socket.recv(1024))
                userpasswd = server_db.get_user_passwd(response['username'])
                if hmac.compare_digest(userpasswd, response['passwd'].encode(encoding='utf-8')):
                    log.info(f'Успешная аутентификация {response["username"]}')
                    self.remote_socket.send('Auth ok!'.encode('utf-8'))
                else:
                    log.info(f'Ошибка при аутентификации пользователя {response["username"]}')
                    self.remote_socket.send('Auth fail!'.encode('utf-8'))

            else:
                # В список "рабочих" сокетов его не добавляем, не слушаем
                log.info(f'Ошибка авторизации {self.remote_socket} {self.address}')
                # Будем вежливыми до конца, сообщим клиенту о неудаче авторизации
                self.remote_socket.send('Auth fail!'.encode('utf-8'))

    def exchange_messages(self):
        """
        Метод, выполняющий прослушивание сокетов на предмет приема/отправки сообщений
        :return:
        """
        wait = 5
        self.read_sockets = []   # Список клиентов, читающих сообщения
        self.write_sockets = []  # Список клиентов, отправляющих сообщения
        try:
            self.read_sockets, self.write_sockets, e = select.select(self.sockets, self.sockets, [], wait)
        except Exception as ex:
            print(f'select exception: {ex}')
            pass  # Ничего не делать, если какой-то клиент отключился

    def read_messages(self):
        """
        Метод читает сообщения из сокетов
        В зависимости от типа принятого сообщения выполняет нужную операцию и формирует ответ
        для дальнейшей отправки
        :return:
        """
        for client_socket in self.write_sockets:
            try:
                client_socket.settimeout(0.1)  # !!! эта строка обязательна!
                data = client_socket.recv(1024)  # Получили сообщение от клиента

            except OSError as e:
                pass  # отвалились по таймауту

            except:
                pass  # не получили сообщение от клиента

            else:
                if data:
                    msg = decode_msg(data)  # Декодировали сообщение от клиента
                    if msg:
                        if msg['action'] == 'append_user':
                            # TODO: вынести в отдельный метод
                            # Добавили username:socket в словарь
                            self.usersdict[msg['from_user']] = client_socket

                            # Здесь уже не создаем нового пользователя, как в ДЗ-5
                            # Ищем зарегистрированного пользователя и добавляем сессию
                            id_user = server_db.get_user_id(msg['from_user'])
                            id_session = server_db.append_session(id_user, self.address[0], self.address[1])
                            self.user_session[msg['from_user']] = id_session

                        elif msg['action'] == 'get_users':
                            # TODO: вынести в отдельный метод
                            # Обрабатываем запрос на получение списка пользователей
                            # UPD: устарело! Было актуально для cli-версии
                            loc_userslist = server_db.get_users_list()
                            self.responses.append({'action': 'userslist',
                                                   'from_user': 'server',
                                                   'to_user': msg['from_user'],
                                                   'message': ', '.join(loc_userslist)})

                        elif msg['action'] == 'get_users_status':
                            # TODO: вынести в отдельный метод
                            # Запрос на получение статуса пользователей (онлайн/офлайн)
                            loc_userslist = msg['userslist'].split(',')
                            loc_online_userslist = []

                            for item in loc_userslist:
                                if server_db.get_user_session(server_db.get_user_id(item)):
                                    loc_online_userslist.append(item)

                            # В ответ отправляем только список online-пользователей
                            self.responses.append({'action': 'get_users_status',
                                                   'from_user': 'server',
                                                   'to_user': msg['from_user'],
                                                   'userslist': ','.join(loc_online_userslist)})

                        elif msg['action'] == 'kill_user':
                            # TODO: вынести в отдельный метод
                            # Прилетел запрос на отключение от {msg["from_user"]}')

                            # Удаляем клиента из self.usersdict
                            del self.usersdict[msg["from_user"]]

                            # Закрываем сессию в Users
                            server_db.close_session(self.user_session[msg['from_user']])

                            # Обнуляем номер сессии в словаре
                            self.user_session[msg['from_user']] = 0

                        elif msg['action'] == 'check_username':
                            # TODO: вынести в отдельный метод
                            # Проверяем существование пользователя на сервере
                            is_user_exists = server_db.check_username(msg['username'])

                            # Формируем ответ
                            self.responses.append({'action': 'check_username',
                                                   'from_user': 'server',
                                                   'to_user': msg['from_user'],
                                                   'username': msg['username'],
                                                   'message': is_user_exists})

                            # Сохраняем сообщение сообщение в БД
                            server_db.append_message(msg['from_user'], msg['to_user'], msg['username'])

                        else:
                            # TODO: вынести в отдельный метод
                            # Обрабатываем обычное сообщение
                            loc_userslist = server_db.get_users_list()
                            if msg['to_user'] in loc_userslist:
                                self.responses.append({'action': 'message',
                                                       'from_user': msg['from_user'],
                                                       'to_user': msg['to_user'],
                                                       'message': msg['message'].upper()})

                                # Сохраняем сообщение в БД
                                server_db.append_message(msg['from_user'], msg['to_user'], msg['message'])
                            else:
                                print(f'В списке активных пользователей не найден {msg["to_user"]}')

    def write_responses(self):
        """
        Метод рассылает накопленные ответы клиентам
        :return:
        """
        for response in self.responses:
            client_socket = self.get_client_socket(response['to_user'])
            if client_socket is None:
                print(f'Клиент {response["to_user"]} отвалился, сокет недоступен')
                # TODO: вынести в отдельный метод
                # Удаляем клиента из self.usersdict
                try:
                    del self.usersdict[response['to_user']]
                except:
                    pass

                # Удаляем из self.sockets
                try:
                    self.sockets.remove(client_socket)
                except:
                    pass

                # Закрываем сессию в Users
                server_db.save_user_end_session(response['to_user'])
            else:
                response = encode_msg(response)
                try:
                    client_socket.send(response)

                except:
                    # Фактически здесь exception не срабатывает. Нужно сделать получение от клиента подтверждения о
                    # получении сообщения. Только тогда можно быть уверенным, что сообщение доставлено
                    # Эта ситуация может возникнуть при нештатном завершении сессии клиента
                    # Сокет недоступен, клиент отвалился
                    log.error('Клиент {} {} отключился'.format(client_socket.fileno(), client_socket.getpeername()))
                    print(f'Клиент {response["to_user"]} отвалился, сокет недоступен')
                    client_socket.close()

                    # Удаляем клиента из self.usersdict
                    del self.usersdict[response['to_user']]

                    # Удаляем из self.sockets
                    self.sockets.remove(client_socket)

                    # Закрываем сессию в Users
                    server_db.save_user_end_session(self.db_session, response['to_user'])
                else:
                    # Сообщение успешно отправлено
                    pass

        # Отправили все сообщения из списка, список очистили
        self.responses.clear()


@log
def get_logger():
    l = None
    try:
        l = logging.getLogger('server')
    except Exception as e:
        print(f'Возникла ошибка при создании логгера {str(e)}')
        sys.exit(1)
    else:
        l.info(f'Логгер успешно создан: {l}')

    return l


if __name__ == '__main__':
    log = get_logger()
    hosts, port, db_filename = get_server_params(sys.argv, 'server')

    server_db = ServerDB(db_filename)

    server = Server()
    server.hosts = hosts
    server.port = port

    while True:
        # Проверяем новые соединения
        server.accept_connection()

        # Ищем "работающие" сокеты
        server.exchange_messages()

        # Читаем сообщения из сокетов
        server.read_messages()  # Получаем сообщения от клиентов

        # Отправляем ответы
        server.write_responses()  # Отправляем полученные сообщения читателям
